"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var either_1 = require("./either");
var chai = require("chai");
var chaiAsPromised = require("chai-as-promised");
chai.use(chaiAsPromised);
var expect = chai.expect;
var anError = function () { return new Error('Something is wrong'); };
var add1 = function (n) { return n + 1; };
var removeFirstElement = function (arr) {
    if (arr.length === 0) {
        throw new Error('Array is empty');
    }
    return arr.slice(1);
};
var pipeline = function (initialValue) {
    var fns = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        fns[_i - 1] = arguments[_i];
    }
    return fns.reduce(function (acc, fn) { return fn(acc); }, initialValue);
};
describe('Either', function () {
    describe('isRight', function () {
        it('should return false when Left is provided', function () {
            expect(either_1.isRight(either_1.left(anError()))).to.be.false;
        });
        it('should return true when Right is provided', function () {
            expect(either_1.isRight(either_1.right('hola'))).to.be.true;
        });
        it('should return false when a non Either type is provided', function () {
            var nonEither = [1, 2, 3];
            expect(either_1.isRight(nonEither)).to.be.false;
        });
        it('should return false when null or undefined is provided', function () {
            expect(either_1.isRight(null)).to.be.false;
            expect(either_1.isRight(undefined)).to.be.false;
        });
    });
    describe('isLeft', function () {
        it('should return false when Right is provided', function () {
            expect(either_1.isLeft(either_1.right('hola'))).to.be.false;
        });
        it('should return true when Left is provided', function () {
            expect(either_1.isLeft(either_1.left(anError()))).to.be.true;
        });
        it('should return true when a non Either type is provided', function () {
            var nonEither = [1, 2, 3];
            expect(either_1.isLeft(nonEither)).to.be.true;
        });
        it('should return true when null or undefined is provided', function () {
            expect(either_1.isLeft(null)).to.be.true;
            expect(either_1.isLeft(undefined)).to.be.true;
        });
    });
    describe('withDefault', function () {
        it('should return the Right value when Right is provided', function () {
            expect(either_1.withDefault(either_1.right(6), 0)).to.equal(6);
        });
        it('should return the default value when Left is provided', function () {
            expect(either_1.withDefault(either_1.left(anError()), 0)).to.equal(0);
        });
        it('should return the default value when a non Either type is provided', function () {
            var nonEither = [1, 2, 3];
            expect(either_1.withDefault(nonEither, 2)).to.eq(2);
        });
        it('should return the default value when null or undefined is provided', function () {
            expect(either_1.withDefault(null, 1)).to.eq(1);
            expect(either_1.withDefault(undefined, 1)).to.eq(1);
        });
    });
    describe('map', function () {
        it('should return Right when mapping over Right', function () {
            var four = either_1.right(4);
            var five = either_1.map(add1, four);
            expect(either_1.isRight(five)).to.be.true;
        });
        it('should return Left when mapping over Left', function () {
            var result = either_1.map(add1, either_1.left(anError()));
            expect(either_1.isLeft(result)).to.be.true;
        });
        it('should return 5 when adding 1 to Right(4)', function () {
            var four = either_1.right(4);
            var five = either_1.map(add1, four);
            expect(either_1.isRight(five)).to.be.true;
            expect(either_1.withDefault(five, 0)).to.equal(5);
        });
        it('should remove the first element of [1,2,3] succesfully', function () {
            var list = either_1.map(removeFirstElement, either_1.right([1, 2, 3]));
            expect(either_1.isRight(list)).to.be.true;
            expect(either_1.withDefault(list, [])).to.eql([2, 3]);
        });
        it('should return a Left when mapping produces an exception', function () {
            var list = either_1.map(removeFirstElement, either_1.right([]));
            expect(either_1.isLeft(list)).to.be.true;
        });
        it('should return a Left when mapping over null, undefined or any non Either type', function () {
            expect(either_1.isLeft(either_1.map(add1, null))).to.be.true;
            expect(either_1.isLeft(either_1.map(add1, undefined))).to.be.true;
            expect(either_1.isLeft(either_1.map(add1, {}))).to.be.true;
        });
    });
    describe('andThen', function () {
        var safeRemoveFirst = function (arr) {
            try {
                return either_1.right(removeFirstElement(arr));
            }
            catch (error) {
                return either_1.left(error);
            }
        };
        it('should perform chained Right transformations', function () {
            var result = either_1.andThen(function (arr) { return safeRemoveFirst(arr); }, safeRemoveFirst(['a', 'b', 'c']));
            expect(either_1.withDefault(result, ['default val'])).to.eql(['c']);
        });
        it('should perform chained Right transformations and fail with Left', function () {
            var result = either_1.andThen(function (arr) { return either_1.andThen(function (arr2) { return safeRemoveFirst(arr2); }, safeRemoveFirst(arr)); }, safeRemoveFirst(['a', 'b']));
            expect(either_1.withDefault(result, ['default val'])).to.eql(['default val']);
        });
        it('should perform chained Right transformations and fail with Left 2', function () {
            var result = pipeline(['a', 'b', 'c'], safeRemoveFirst, function (arr) { return either_1.andThen(safeRemoveFirst, arr); }, function (arr) { return either_1.andThen(safeRemoveFirst, arr); }, function (arr) { return either_1.andThen(safeRemoveFirst, arr); }, function (arr) { return either_1.andThen(safeRemoveFirst, arr); }, function (arr) { return either_1.andThen(safeRemoveFirst, arr); }, function (arr) { return either_1.andThen(safeRemoveFirst, arr); });
            expect(either_1.caseOf({
                Right: function (arr) { return arr.toString(); },
                Left: function (err) { return err.message; }
            }, result)).to.equal('Array is empty');
        });
        it('should return left throw when chaining over null, undefined or any non Either type', function () {
            expect(either_1.isLeft(either_1.andThen(function (a) { return either_1.right(a); }, null))).to.be.true;
            expect(either_1.isLeft(either_1.andThen(function (a) { return either_1.right(a); }, undefined))).to.be.true;
            expect(either_1.isLeft(either_1.andThen(function (a) { return either_1.right(a); }, {}))).to.be.true;
        });
    });
    describe('caseOf', function () {
        it('should Launch 5 missiles', function () {
            expect(either_1.caseOf({
                Left: function (err) { return "Error: " + err.message; },
                Right: function (n) { return "Launch " + n + " missiles"; }
            }, either_1.right('5'))).to.equal('Launch 5 missiles');
        });
        it('should error', function () {
            expect(either_1.caseOf({
                Left: function (err) { return "Error: " + err.message; },
                Right: function (n) { return "Launch " + n + " missiles"; }
            }, either_1.left(anError()))).to.equal('Error: Something is wrong');
        });
    });
    describe('tryCatch', function () {
        var shouldFail;
        var fn = function () {
            if (shouldFail) {
                throw anError();
            }
            return 'Ok';
        };
        it('should convert a failing function into a Left', function () {
            shouldFail = true;
            expect(either_1.caseOf({
                Left: function (err) { return "Error: " + err.message; },
                Right: function (n) { return "Launch " + n + " missiles"; }
            }, either_1.tryCatch(fn, function (err) { return err; }))).to.equal('Error: Something is wrong');
        });
        it('should convert a successful function into a Right', function () {
            shouldFail = false;
            expect(either_1.tryCatch(fn, function (err) { return err; })).to.deep.equal(either_1.right('Ok'));
        });
    });
    describe('example', function () {
        it('should work', function () {
            var readFile = function (filename) {
                var fileSystem = {
                    'something.json': "\n          [\n            {\n              \"id\": 1,\n              \"nickname\": \"rick\",\n              \"email\": \"rick@c137.com\",\n              \"bio\": \"Rick Sanchez of Earth Dimension C-137\",\n              \"dob\": \"3139-03-04T23:00:00.000Z\"\n            },\n            {\n              \"id\": 2,\n              \"nickname\": \"morty\",\n              \"email\": \"morty@c137.com\",\n              \"bio\": null,\n              \"dob\": \"2005-04-08T22:00:00.000Z\"\n            }\n          ]"
                };
                var fileContents = fileSystem[filename];
                if (fileContents === undefined) {
                    throw new Error(filename + " does not exists.");
                }
                return fileContents;
            }; // throws if file does not exists
            var getUserById = function (filename, id) {
                var validateJsonFilename = function (filename) {
                    return filename.endsWith('.json')
                        ? either_1.right(filename)
                        : either_1.left(new Error(filename + " is not a valid json file."));
                };
                var readFileContent = function (filename) {
                    return either_1.tryCatch(function () { return readFile(filename); }, function (err) { return err; });
                };
                var parseJson = function (json) {
                    return either_1.tryCatch(function () { return JSON.parse(json); }, function (err) { return new Error("There was an error parsing this Json."); });
                };
                var findUserById = function (users) {
                    return either_1.right(users.find(function (user) { return user.id === id; }) || null);
                };
                return pipeline(filename, function (fname) { return validateJsonFilename(fname); }, function (fname) { return either_1.andThen(readFileContent, fname); }, function (json) { return either_1.andThen(parseJson, json); }, function (users) { return either_1.andThen(findUserById, users); });
            };
            expect(getUserById('something.json', 2)).to.deep.equal(either_1.right({
                id: 2,
                nickname: 'morty',
                email: 'morty@c137.com',
                bio: null,
                dob: '2005-04-08T22:00:00.000Z'
            }));
            expect(getUserById('something.json', 3)).to.deep.equal(either_1.right(null));
            expect(either_1.caseOf({
                Left: function (err) { return "Error: " + err.message; },
                Right: function (n) { return "Launch " + n + " missiles"; }
            }, getUserById('nothing.json', 999))).to.equal('Error: nothing.json does not exists.');
        });
        it('should work simple', function () {
            var readFile = function (filename) {
                var fileSystem = {
                    'something.json': "\n          [\n            {\n              \"id\": 1,\n              \"nickname\": \"rick\",\n              \"email\": \"rick@c137.com\"\n            },\n            {\n              \"id\": 2,\n              \"nickname\": \"morty\",\n              \"email\": \"morty@c137.com\"\n            }\n          ]"
                };
                var fileContents = fileSystem[filename];
                if (fileContents === undefined) {
                    throw new Error(filename + " does not exists.");
                }
                return fileContents;
            }; // throws if file does not exists
            var readFileContent = function (filename) {
                return either_1.tryCatch(function () { return readFile(filename); }, function (err) { return err; });
            };
            var parseJson = function (json) {
                return either_1.tryCatch(function () { return JSON.parse(json); }, function (err) { return new Error("There was an error parsing this Json."); });
            };
            var usersFull = pipeline('something.json', function (fname) { return readFileContent(fname); }, function (json) { return either_1.andThen(parseJson, json); }, function (users) { return either_1.withDefault(users, []); });
            var usersEmpty = pipeline('nothing.json', function (fname) { return readFileContent(fname); }, function (json) { return either_1.andThen(parseJson, json); }, function (users) { return either_1.withDefault(users, []); });
            expect(usersFull).to.deep.equal([
                {
                    id: 1,
                    nickname: 'rick',
                    email: 'rick@c137.com'
                },
                {
                    id: 2,
                    nickname: 'morty',
                    email: 'morty@c137.com'
                }
            ]);
            expect(usersEmpty).to.deep.equal([]);
        });
    });
});
